"""
RelationDAG
"""

from functools import lru_cache
from typing import Iterable, List, Mapping, NamedTuple, Tuple, Type

import matplotlib.pyplot as plot
import networkx
from loguru import logger
from networkx import DiGraph
from sqlalchemy import ForeignKey, MetaData
from sqlalchemy.schema import Column, Table
from sqlalchemy_utils import get_referencing_foreign_keys

from scenographer.database import Database, UUIDField


class Relation(NamedTuple):
    "Container for the associated pair of primary key and foreign key"
    pk: Column
    fk: Column

    @property
    def edge(self) -> Tuple[Table, Table, Mapping[str, Column]]:
        "Returns a tuple that is accepted by networkx lib as a graph edge"
        return (self.pk.table, self.fk.table, {"relation": self})

    def __repr__(self) -> str:
        "REPL representation of a relation"
        relation_kwargs = [
            f"pk=database_tables.{self.pk.table.name}.c.{self.pk.name}",
            f"fk=database_tables.{self.fk.table.name}.c.{self.fk.name}",
        ]

        return f"Relation({','.join(relation_kwargs)})"

    @classmethod
    def from_foreign_key(cls: Type["Relation"], foreign_key: ForeignKey) -> "Relation":
        "Constructor for a relation given a ForeignKey object"
        return cls(
            pk=foreign_key.column,  # referenced column
            fk=foreign_key.parent,  # referer column
        )

    @classmethod
    def from_tables(cls: Type["Relation"], tables: List[Table]) -> List["Relation"]:
        """
        Constructor for a list of relations given a table
        Creates one relation for each foreign_key constraint associated with the table
        """
        return [
            cls.from_foreign_key(fk)
            for table in tables
            for fk in get_referencing_foreign_keys(table)
        ]


class RelationDAG(NamedTuple):
    "Wrapper for operations around the graph of relations"
    graph: DiGraph

    @property
    @lru_cache()
    def tables(self) -> List[Table]:
        "Lists all tables taken into consideration for sampling"
        return self.graph.nodes

    @property
    @lru_cache()
    def entrypoints(self) -> List[Table]:
        "Lists tables which have no foreign keys"
        return [n for (n, d) in self.graph.in_degree(self.graph.nodes) if d == 0]

    @property
    @lru_cache()
    def topologically_sorted(self) -> Iterable[Table]:
        """
        Returns all tables ordered in a way that
        if table X has a foreign key to Y, Y will always come first
        """
        return networkx.topological_sort(self.graph)

    def write_plot(self, filepath: str = "graph.png") -> None:
        "Meh graph image representation"
        logger.debug("Writing image file with graph")
        plot.title("RelationDAG")
        pos = networkx.drawing.nx_agraph.graphviz_layout(self.graph, prog="dot")
        plot.figure(figsize=(40, 40))

        networkx.draw_networkx_nodes(self.graph, pos, node_size=5000, alpha=0.9)
        networkx.draw_networkx_edges(
            self.graph, pos, arrows=True, node_size=5000,
        )
        networkx.draw_networkx_labels(self.graph, pos)

        plot.axis("off")
        plot.savefig(filepath)

    def write_dot(self, filepath: str = "graph.dot") -> None:
        "Meh graph image representation. File must be rendered with `dot`"
        logger.debug("Writing dot file with graph")
        networkx.drawing.nx_agraph.write_dot(self.graph, filepath)

    @property
    @lru_cache()
    def key_schema(self) -> MetaData:
        """
        Create schema with only the primary keys and foreign keys of each table.
        Ensures that the resulting schema is compatible with sqlite3
        """
        metadata = MetaData()

        for table in self.graph.nodes:

            # Here we use edge data for the first (only ?) time.
            # Perhaps we can shape the data better to avoid doing work here.

            # Select the relations whose foreign keys are present in this table
            relations = [
                edge[-1]["relation"]
                for edge in self.graph.reverse(copy=False).edges(
                    nbunch=table, data=True
                )
            ]

            # We get PK from table data instead of relation data,
            # because a primary key doesn't necessarily form a relation.
            # Assumes PK either is a single column or it doesn't exist.
            primary_key = [
                Column(c.name, c.type, primary_key=True)
                for c in table.primary_key.columns
            ][:1]

            columns = [
                *primary_key,
                *[
                    Column(r.fk.name, r.fk.type, ForeignKey(str(r.pk)))
                    for r in relations
                ],
            ]

            # Map postgres UUIDs into sqlite compatible UUIDs
            for column in columns:
                if str(column.type) not in ["UUID", "BIGINT", "INTEGER"]:
                    logger.warning(
                        "Key column {} has weird type {}",
                        f"{table.name}.{column.name}",
                        str(column.type),
                    )
                if str(column.type) == "UUID":
                    column.type = UUIDField()

            Table(table.name, metadata, *columns)

        return metadata

    @classmethod
    def from_graph(cls: Type["RelationDAG"], graph: DiGraph) -> "RelationDAG":
        """
        Instanciates a RelationDAG from a networkx DiGraph.
        It makes the graph is immutable and raises if the graph is not a DAG
        """
        new = cls(networkx.freeze(graph))
        if networkx.is_directed_acyclic_graph(new.graph):
            logger.debug(
                "DAG contains {} nodes and {} edges",
                len(graph.nodes),
                len(graph.edges),
            )
        else:
            logger.error("Generated graph is not a DAG.")
            raise ValueError

        return new

    @classmethod
    def load(
        cls: Type["RelationDAG"],
        database: Database,
        extend_relations: List[Relation],
        ignore_relations: List[Relation],
        ignore_tables: List[Table],
    ) -> "RelationDAG":
        """
        Create a RelationDAG
        The data loaded from this method is sourced
        from the database and from the user config
        """
        graph = DiGraph(name="RelationDAG")

        # Get actual table instances
        tables = database.tables.__dict__.values()

        # Create relations from table data and add the ones specified in settings
        relations = Relation.from_tables(tables) + extend_relations

        # Create graph
        graph.add_nodes_from(tables)
        graph.add_edges_from([r.edge for r in relations])

        # Remove excluded entities (tables and relations) from the created graph
        graph.remove_edges_from([r.edge for r in ignore_relations])
        graph.remove_nodes_from(ignore_tables)

        # Create RelationDAG instance
        return cls.from_graph(graph)

    def __str__(self) -> str:
        "Return some useful information about the graph"
        return networkx.info(self.graph)

    def __repr__(self) -> str:
        "For REPL use. Should work with only RelationDAG in scope."
        return "RelationDAG"



"""
This module takes care of loading the settings, transforming them
into an appropriate format if necessary and providing the
user-provided settings to the rest of the application.
"""

import json
import operator
import os
import sys
from functools import lru_cache
from pathlib import Path
from tempfile import mkdtemp
from typing import Any, Iterable, List, Mapping, NamedTuple, Optional

import sqlalchemy
from dict_digger import dig
from loguru import logger
from pyrsistent import freeze, thaw
from sqlalchemy.schema import Table

from scenographer.database import Database
from scenographer.relations import Relation, RelationDAG
from scenographer.sample import TableSample


class Settings(NamedTuple):
    SOURCE_DATABASE_URL: str
    TARGET_DATABASE_URL: str
    QUERY_MODIFIERS: dict = freeze({})
    IGNORE_TABLES: List[str] = freeze([])
    EXTEND_RELATIONS: List[Mapping[str, str]] = freeze([])
    IGNORE_RELATIONS: List[Mapping[str, str]] = freeze([])
    OUTPUT_DIRECTORY: Optional[str] = None

    @classmethod
    def load(cls, path):
        import commentjson

        with open(path) as f:
            data = commentjson.load(f)

        return cls(**freeze(data))

    @classmethod
    @lru_cache()
    def empty(cls):
        return cls(
            SOURCE_DATABASE_URL="",
            TARGET_DATABASE_URL="",
            IGNORE_TABLES=freeze(["example1", "migrations"]),
            EXTEND_RELATIONS=freeze(
                [{"pk": "product.id", "fk": "product_ownership.product_id"}]
            ),
            IGNORE_RELATIONS=freeze(
                [{"pk": "product.id", "fk": "client.favorite_product_id"}]
            ),
            QUERY_MODIFIERS=freeze(
                {
                    "_default": {"conditions": [], "limit": 300},
                    "users": {"conditions": ["email ilike '%@example.com'"]},
                }
            ),
        )

    @property
    @lru_cache()
    def json(self):
        return json.dumps(
            {k: thaw(v) for k, v in self._asdict().items()}, indent=4, sort_keys=True,
        )


class Scenographer(NamedTuple):
    options: Settings

    @property
    @lru_cache()
    def source_database(self) -> Database:
        """
        SOURCE_DATABASE_URL key.
        Abort if any any issue arises!
        """
        return self._instanciate_database("SOURCE_DATABASE_URL")

    @property
    def target_database(self) -> Database:
        """
        TARGET_DATABASE_URL key.
        Abort if any any issue arises!
        """
        return self._instanciate_database("TARGET_DATABASE_URL")

    @property
    @lru_cache()
    def output_directory(self) -> Path:
        """
        SOURCE_DATABASE_URL key.
        Abort if any any issue arises!
        """
        if not self.options.OUTPUT_DIRECTORY:
            directory = mkdtemp(prefix="sample-")
            logger.info(
                "Output directory unspecified. "
                "Saving samples in temporary directory {}",
                directory,
            )
        else:
            directory = self.options.OUTPUT_DIRECTORY
            try:
                os.mkdir(directory)
            except FileExistsError:
                logger.error(
                    "Error creating output directory {}. "
                    "Make sure there's nothing there. "
                    "Aborting.",
                    directory,
                )
                sys.exit(1)

        return Path(directory)

    @property
    @lru_cache()
    def extend_relations(self) -> Iterable[Relation]:
        """
        EXTEND_RELATIONS key.
        Scenographer will recognize the relations specified here,
        in addition to those specified by database constraints.
        This method will return the equivalent Relation instances
        """
        relations = []
        for relation in self.options.EXTEND_RELATIONS:
            try:
                relations.append(
                    Relation(
                        pk=self.column_by_name[relation["pk"]],
                        fk=self.column_by_name[relation["fk"]],
                    )
                )
            except KeyError:
                logger.warning("Can't match relation {}. Skipping.", dict(relation))
                continue

        return relations

    @property
    @lru_cache()
    def ignore_relations(self) -> Iterable[Relation]:
        """
        IGNORE_RELATIONS key.
        Scenographer will not recognize the relations specified here.
        This method will return the equivalent Relation instances
        """
        relations = []
        for relation in self.options.IGNORE_RELATIONS:
            try:
                relations.append(
                    Relation(
                        pk=self.column_by_name[relation["pk"]],
                        fk=self.column_by_name[relation["fk"]],
                    )
                )
            except KeyError:
                logger.warning("Can't match relation {}. Skipping.", relation)
                continue

        return relations

    @property
    @lru_cache()
    def ignore_tables(self) -> Iterable[Table]:
        """
        IGNORE_TABLES key.
        Scenographer will not sample any data coming from these tables.
        This method will return the equivalent Table instances
        (binded to source_database)
        """
        ignored_tables = []
        for table_name in self.options.IGNORE_TABLES:
            if table_name in self.source_database.tables.__dict__:
                ignored_tables.append(getattr(self.source_database.tables, table_name))
            else:
                logger.warning("Can't find table {}. Skipping.", table_name)
        return ignored_tables

    @property
    @lru_cache()
    def query_modifiers(self) -> Mapping[Table, Mapping[str, Any]]:
        """
        QUERY_MODIFIERS key.
        Scenographer will take this key in account while sampling.
        This method will return the equivalent Table instances
        (binded to source_database)
        """
        mods = self.options.QUERY_MODIFIERS
        default_limit = dig(thaw(mods), "_default", "limit") or 30
        default_conditions = dig(thaw(mods), "_default", "conditions") or list()

        non_specified_entrypoints = operator.sub(
            set([t.name for t in self.relation_dag.entrypoints]), set(mods.keys()),
        )
        if non_specified_entrypoints:
            logger.warning(
                "Entrypoints are advised to be added as query modifiers. "
                "They define what the final sample will look like"
            )
            logger.warning(
                "These entrypoints are not specified: {}", non_specified_entrypoints,
            )

        modifiers = {}
        for table in self.relation_dag.tables:
            if table.name not in mods:
                limit = default_limit
                conditions = default_conditions

            else:
                limit, conditions = (
                    mods[table.name].get("limit"),
                    mods[table.name].get("conditions"),
                )
                if not limit and not conditions:
                    logger.warning("QUERY_MODIFIER for {} malformed.", table.name)
                    continue

                limit = limit or default_limit
                conditions = conditions or default_conditions

            modifiers[table] = {"limit": limit, "conditions": conditions}

        return modifiers

    @property
    @lru_cache()
    def relation_dag(self) -> RelationDAG:
        logger.info("Building the database graph")
        return RelationDAG.load(
            self.source_database,
            extend_relations=self.extend_relations,
            ignore_relations=self.ignore_relations,
            ignore_tables=self.ignore_tables,
        )

    @property
    def samples(self) -> Iterable[Table]:
        logger.info("Starting sampling")
        return TableSample.sample_dag(
            self.source_database,
            self.relation_dag,
            self.query_modifiers,
            self.output_directory,
        )

    @property
    @lru_cache()
    def column_by_name(self):
        return {
            f"{column.table.name}.{column.name}": column
            for columns in [
                table.columns for table in self.source_database.tables.__dict__.values()
            ]
            for column in columns
        }

    def copy_schema(self):
        logger.info("Copying schema")
        self.target_database.load_schema(self.source_database)
        return self

    def copy_sample(self):
        logger.info("Loading sample into target")
        self.target_database.load_samples(self.output_directory, self.samples)
        return self

    def _instanciate_database(self, var):
        """
        Helper method to instanciate a database method
        Scenographer will read a postgres url from the config
        and try to connect with that database.
        Abort if any any issue arises!
        """
        try:
            database_url = self.options._asdict()[var]
            return Database(database_url).test_conn()
        except sqlalchemy.exc.ArgumentError:
            logger.error("Error connecting to {}", var)
            sys.exit(1)



#!/usr/bin/env python3

"""
Module to connect with database
"""

import subprocess
import uuid
from functools import lru_cache
from pathlib import Path
from pipes import quote
from types import SimpleNamespace
from typing import Any, Iterable, List, Mapping, NamedTuple, TextIO

import postgres_copy
import sqlalchemy
from loguru import logger
from sqlalchemy.dialects.postgresql import UUID
from sqlalchemy.engine import Engine, ResultProxy
from sqlalchemy.schema import Table
from sqlalchemy.sql.expression import Select
from sqlalchemy.types import CHAR, TypeDecorator

from scenographer.utils import PrintAs


class Database(NamedTuple):
    "Wrapper around database operations"
    database_url: str

    def execute(self, *args, **kwargs) -> ResultProxy:
        """
        Executes query and returns the usual ResultProxy from SQLAlchemy.
        """
        with self.engine.begin() as connection:
            return connection.execute(*args, **kwargs)

    def execute_return_list(self, *args, **kwargs) -> List[Any]:
        """
        Executes query and returns a list of the resulting values
        It raises an AssertionError if more than one column is returned
        """
        resultproxy = list(self.execute(*args, **kwargs))

        if resultproxy:
            columns = [column_name for column_name, value in resultproxy[0].items()]
            if len(columns) != 1:
                print(columns)
            assert len(columns) == 1

        return [rowproxy.values()[0] for rowproxy in resultproxy]

    def execute_return_dict(self, *args, **kwargs) -> List[Mapping[str, Any]]:
        """
        Executes query and returns each row as a dictionary;
        It raises an AssertionError if any column name is repeated.
        """
        resultproxy = list(self.execute(*args, **kwargs))

        if resultproxy:
            columns = [column_name for column_name, value in resultproxy[0].items()]
            assert len(columns) == len(set(columns))

        return [
            {column: value for column, value in rowproxy.items()}
            for rowproxy in resultproxy
        ]

    def copy_to_csv(self, file: TextIO, select: Select) -> None:
        "Executes query and write the rows into a file object in CSV format."
        postgres_copy.copy_to(select, file, self.engine, format="csv", header=True)

    @property
    @lru_cache()
    def engine(self) -> Engine:
        "Create, return and cache the associated sqlalchemy engine"
        return sqlalchemy.create_engine(self.database_url)

    @property
    @lru_cache()
    def tables(self) -> SimpleNamespace:
        "Reflect the database to return and cache a namespace with all of its tables"

        logger.info("Reflecting source database")
        metadata = sqlalchemy.MetaData()
        with PrintAs(logger.warning):
            metadata.reflect(self.engine, views=False)

        return SimpleNamespace(**metadata.tables)

    def load_schema(self, source_database: "Database") -> None:
        """
        pg_dump \
            --format=custom --no-owner --schema-only \
            --verbose {source_database} \
        | pg_restore \
            --format=custom --no-owner --schema-only \
            --no-acl \
            --verbose -d {target_database}
        """
        pg_copy_schema = self.load_schema.__doc__.format(
            source_database=quote(source_database.database_url),
            target_database=quote(self.database_url),
        )
        process = subprocess.Popen(
            pg_copy_schema,
            shell=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
        )
        with process.stdout:
            for line in iter(process.stdout.readline, b""):
                logger.trace("{}", line)

        exit_code = process.wait()
        logger.debug("Command pg_dump | pg_restore exited with code {}", exit_code)
        # assert exit_code == 0

    def load_samples(self, directory: Path, samples: Iterable[Table]) -> None:
        "Copy the generated sample CSVs into the database"
        for table in samples:
            with open(directory / Path(table.name).with_suffix(".csv")) as file:
                postgres_copy.copy_from(
                    file, table, self.engine, format="csv", header=True
                )

    def test_conn(self) -> "Database":
        "Copy the generated sample CSVs into the database"
        return self.engine and self  # Always returns self


# Retrieved from
# https://docs.sqlalchemy.org/en/13/core/custom_types.html#backend-agnostic-guid-type
class UUIDField(TypeDecorator):
    """Platform-independent GUID type.

    Uses PostgreSQL's UUID type, otherwise uses
    CHAR(32), storing as stringified hex values.
    """

    impl = CHAR

    def load_dialect_impl(self, dialect):
        if dialect.name == "postgresql":
            return dialect.type_descriptor(UUID())
        else:
            return dialect.type_descriptor(CHAR(32))

    def process_bind_param(self, value, dialect):
        if value is None:
            return value
        elif dialect.name == "postgresql":
            return str(value)
        else:
            if not isinstance(value, uuid.UUID):
                return "%.32x" % uuid.UUID(value).int
            else:
                # hexstring
                return "%.32x" % value.int

    def process_result_value(self, value, dialect):
        if value is None:
            return value
        else:
            if not isinstance(value, uuid.UUID):
                value = uuid.UUID(value)
            return value



#!/usr/bin/env python


"""
Scenographer - The cool dude who sets up the stage. Word.

Usage:
  scenographer sample <config-path> [--skip-schema]
  scenographer empty-config
  scenographer -h | --help
  scenographer --version

Options:
  -h --help                Show this screen.
  --version                Show version.
"""

from docopt import docopt

from scenographer.scenographer import Scenographer, Settings


def cli():
    args = docopt(__doc__, version="0.2.0")

    if args["sample"] is True:
        scenographer = Scenographer(options=Settings.load(args["<config-path>"]))

        if args["--skip-schema"] is False:
            scenographer.copy_schema()

        scenographer.copy_sample()

    elif args["empty-config"]:
        print(Settings.empty().json)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['scenographer']

package_data = \
{'': ['*']}

install_requires = \
['commentjson>=0.8.3,<0.9.0',
 'dict-digger>=0.2.1,<0.3.0',
 'docopt>=0.6.2,<0.7.0',
 'loguru>=0.5.1,<0.6.0',
 'matplotlib>=3.1.2,<4.0.0',
 'networkx>=2.5,<3.0',
 'psycopg2-binary>=2.8.5,<3.0.0',
 'pyrsistent>=0.16.0,<0.17.0',
 'sqlalchemy-postgres-copy>=0.5.0,<0.6.0',
 'sqlalchemy>=1.3.19,<2.0.0',
 'sqlalchemy_utils>=0.36.8,<0.37.0']

entry_points = \
{'console_scripts': ['scenographer = scenographer.cli:cli']}

setup_kwargs = {
    'name': 'scenographer',
    'version': '0.1.1',
    'description': 'The cool dude who sets up the stage. Word.',
    'long_description': '# scenographer\n\n[![PyPI pyversions](https://img.shields.io/pypi/pyversions/scenographer.svg?style=flat-square)](https://pypi.python.org/pypi/scenographer/)\n[![GitHub license](https://img.shields.io/github/license/zyperco/scenographer.svg?style=flat-square)](https://github.com/zyperco/scenographer/blob/master/LICENSE)\n[![PyPI version shields.io](https://img.shields.io/pypi/v/scenographer.svg?style=flat-square)](https://pypi.python.org/pypi/scenographer/)\n\n[![Maintenance](https://img.shields.io/badge/Maintained%3F-yes-green.svg?style=flat-square)](https://GitHub.com/zyperco/scenographer/graphs/commit-activity)\n[![zyperco](https://circleci.com/gh/zyperco/scenographer.svg?style=svg)](https://app.circleci.com/pipelines/github/zyperco/scenographer)\n\n**scenographer** is a Python script that can create a subset of a postgres database, without losing referential integrity.\n\nThe goal is to be able to spawn data-correct databases to easily create new environments that can be used for testing and / or demo\'ing.\n\nRelevant links:\n  - [Documentation](https://zyperco.github.io/scenographer/)\n\n## Installation\n\nUse [pip](https://pip.pypa.io/en/stable/) to install `scenographer`.\n\n```bash\npip install scenographer\n```\n\n## Usage\n\nScenographer requires a configuration file. An empty one, to serve as a starting point, is available by running `scenographer empty-config`.\n\nAfter adjusting the configuration file, it\'s easy to start the sampling run:\n\n```bash\nscenographer bin/scenographer sample config.json\n```\n\nor if the schema doesn\'t need to be recreated in the target database:\n\n```bash\nscenographer bin/scenographer sample config.json --skip-schema\n```\n\n## Configuration\n\n### SOURCE_DATABASE_URL\n\nThe connection string for the source database. Only Postgres is supported.\n\n### TARGET_DATABASE_URL\n\nThe connection string for the target database. Only Postgres is supported.\n\n### IGNORE_RELATIONS\n\nScenographer works by traversing a DAG graph created from the foreign key constraints of the database.\nHowever, it\'s not always the case that the database forms a DAG. To handle those cases, some foreign keys can be ignored by adding exceptions in this form:\n\n```python\nIGNORE_RELATIONS = [\n  {"pk": "product.id", "fk": "client.favorite_product_id"}\n]\n```\n\n### EXTEND_RELATIONS\n\nIn other ocasions, the actual foreign key constraint is not present in the database, although it exists in the business-side of things (like Rails does it).\nAdditional relations can be added to handle those cases. The relations take the same format of `IGNORE_RELATIONS `.\n\n### IGNORE_TABLES\n\nSome tables are _extra_. They may not matter, they may require a special solution or they are part of different components. Either way, you can ignore them.\n\n### QUERY_MODIFIERS\n\nFor some cases, it\'s useful to tap into the actual queries being made. For that, you can add an entry here. Here\'s an example:\n\n```python\nQUERY_MODIFIERS={\n    "_default": {"conditions": [], "limit": 300},\n    "users": {"conditions": ["email ilike \'%@example.com\'"]},\n}\n```\n\nEach entry is a table, with the exception of `_default` which is applied to all queries. Its values can have a `conditions` and/or `limit` key. For conditions you can write plain `sql`.\n\n\n### OUTPUT_DIRECTORY\n\nAt some point, the data is converted into CSV files to be imported into postgres. This is the directory for said CSV files. If you don\'t care about it, feel free to ignore. If it\'s not declared, it will create and use a temporary dictory instead.\n\n\n## Contributing\n\nPull requests are welcome. For major changes, please open an issue first to discuss what you would like to change.\n\nPlease make sure to update tests as appropriate.\n\n## License\n[MIT](https://choosealicense.com/licenses/mit/)\n',
    'author': 'Xavier Francisco',
    'author_email': 'xavier@zyper.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://zyperco.github.io/scenographer/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



import sys
import warnings
from io import StringIO
from typing import Callable


class PrintAs:
    def __init__(self, print_callable: Callable[[str], None], skip_empty=True):
        self.log_sink = StringIO()
        self.print_callable = print_callable
        if skip_empty:
            self._callable = self.print_skip_empty_callable
        else:
            self._callable = self.print_callable

    def __enter__(self):
        self.original_stdout = sys.stdout
        self.original_showwarning = warnings.showwarning
        sys.stdout = self.log_sink
        warnings.showwarning = lambda message, c, f, ln, fh, l: self._callable(message)

    def __exit__(self, exc_type, exc_val, exc_tb):
        sys.stdout.seek(0)
        self._callable(sys.stdout.read())
        sys.stdout.close()
        sys.stdout = self.original_stdout
        warnings.showwarning = self.original_showwarning

    def print_skip_empty_callable(self, message):
        message = str(message).strip()
        message = message.strip()
        if message:
            return self.print_callable(message)



"TableSample"

import os
import sys
from csv import DictReader, field_size_limit
from functools import lru_cache
from pathlib import Path
from typing import Any, Iterable, List, NamedTuple

from loguru import logger
from pyrsistent import freeze
from sqlalchemy import select, text
from sqlalchemy.dialects import postgresql
from sqlalchemy.schema import Column, Table
from sqlalchemy.sql.expression import Select

from scenographer.database import Database
from scenographer.relations import RelationDAG


class TableSample(NamedTuple):
    """
    TableSample is a wrapper around the operations
    to sample a single table
    """

    key_database = Database("sqlite://")

    table: Table
    conditions: Iterable[str]
    limit: int

    @classmethod
    def sample_dag(
        cls,
        source_database: Database,
        relations: RelationDAG,
        query_modifiers: Any,
        directory: Path,
    ) -> List["TableSample"]:
        "Samples a database according to its relation graph"

        # Change current working dir to our output directory
        # This way we avoid TableSample instances containing that information
        os.chdir(directory)

        # Prepare the sqlite db
        relations.key_schema.drop_all(cls.key_database.engine)
        relations.key_schema.create_all(cls.key_database.engine)

        samples = []
        for table in relations.topologically_sorted:
            sample = cls(
                table,
                conditions=freeze(query_modifiers[table]["conditions"]),
                limit=query_modifiers[table]["limit"],
            ).sample(source_database)

            samples.append(sample.table)
        return samples

    # @classmethod
    # @lru_cache()
    # def key_database(cls, relations) -> Database:
    #     return Database("sqlite://")

    @property
    @lru_cache(None)
    def key_table(self) -> Table:
        """
        Returns the table with the associated metadata
        tied to the key database
        """
        return getattr(self.key_database.tables, self.table.name)

    @property
    @lru_cache(None)
    def source_table(self) -> Table:
        """
        Returns the table with the associated metadata
        tied to the source database
        """
        return self.table

    @property
    @lru_cache(None)
    def key_columns(self) -> Iterable[Column]:
        """
        Property defining the columns of the key database:
        the primary_keys and foreign_keys of each table
        """
        return self.key_table.columns

    @property
    @lru_cache(None)
    def foreign_keys(self) -> Iterable[Column]:
        """
        Returns the foreign keys of the table by listing
        the key columns and excluding any primary_key
        """
        return [c for c in self.key_columns if not c.primary_key]

    @property
    @lru_cache(None)
    def is_entrypoint(self) -> bool:
        """
        Property defining if the table has foreign_keys.
        If it is false it means it has no dependencies.
        """
        return not bool(list(self.foreign_keys))

    @property
    @lru_cache(None)
    def query(self) -> Select:
        "SQLAlchemy query object for the sampling query"
        query = self.table.select()
        for condition in self.conditions:
            query = query.where(text(condition))

        if not self.is_entrypoint:
            query = self.follow_conditions(query)

        query = query.limit(self.limit)

        return query

    def follow_conditions(self, query: Select) -> Select:
        "Append WHERE clauses to restrict the rows depending on the data already extracted"
        foreign_key: Column
        primary_key: Column

        # For each foreign_key, build a where clause in the form of:
        #   WHERE sometable_id in (<select id from sometable>)
        # The subquery is executed on the key database and its results are
        # parametrized into the final query.
        for foreign_key in self.foreign_keys:

            primary_key = iter(foreign_key.foreign_keys).__next__().column
            pk_data = self.key_database.execute_return_list(select([primary_key]))
            source_fk = getattr(self.source_table.c, foreign_key.name)

            if pk_data:
                query = query.where(
                    source_fk.in_(pk_data) | source_fk.is_(None)
                ).order_by(source_fk.nullslast())
            else:
                query = query.where(source_fk.is_(None))

        return query

    @property
    @lru_cache(None)
    def sql(self) -> str:
        "Raw SQL representation of the sampling query"
        complete_query = self.query.compile(
            dialect=postgresql.dialect(), compile_kwargs={"literal_binds": True}
        ).__str__()

        # Sampling requires we always extract all the columns, but
        # complete_query comes with the entire columns list.
        # So let's replace the column list with a wildcard:
        query_str = "SELECT * " + complete_query[complete_query.index("FROM") :]
        return query_str

    @property
    @lru_cache(None)
    def persisted_records_path(self) -> Path:
        """
        The path for the csv file for the data to be written in
        Relative to classmethod `sample_dag` directory argument.
        """
        return Path(self.table.name).with_suffix(".csv")

    def persist_keys(self) -> None:
        "Insert retrieved keys from the table csv file into the sqlite key database."
        with self.persisted_records_path.open(newline="") as csv_file:
            keys = [
                {key.name: persisted_record[key.name] for key in self.key_columns}
                for persisted_record in DictReader(csv_file)
            ]
            logger.debug("Got {} records for {}.", len(keys), self.table.name)

        if keys:
            self.key_database.execute(self.key_table.insert(), keys)

    def persist_records(self, source_database) -> None:
        "Insert retrieved rows into persisted_records_path"
        logger.trace("Querying {}\n{}", self.table.name, self.sql)

        with self.persisted_records_path.open("a+") as csv_file:
            source_database.copy_to_csv(csv_file, self.query)

    def sample(self, source_database):
        "Samples the table"
        self.persist_records(source_database)
        self.persist_keys()

        return self


def increase_csv_limit():
    """
    This is needed, due to default limits.
    Taken from https://stackoverflow.com/a/15063941
    """
    max_int = sys.maxsize

    while True:
        # decrease the max_int value by factor 10
        # as long as the OverflowError occurs.

        try:
            field_size_limit(max_int)
            break
        except OverflowError:
            max_int = int(max_int / 10)


increase_csv_limit()
